
from collections import defaultdict
from pprint import pprint
from statistics import median
from typing import Dict, List, Union

from brownie import chain
from pandas import DataFrame

FunctionName = str
Inputs = str
Duration = float

num_recorded_durations = 0
function_durations: Dict[FunctionName,List[Dict[str,Union[Inputs,Duration]]]] = defaultdict(list)


def print_average_durations() -> None:
    average_durations = {}
    for function_name in function_durations:
        _average_duration = average_duration(function_name)
        average_durations[_average_duration] = function_name
    pprint(sorted(average_durations))

def durations(function_name: str) -> List[float]:
    return [event['duration'] for event in function_durations[function_name]]

def average_duration(function_name: str) -> float:
    _durations = durations(function_name)
    return sum(_durations) / len(_durations)

def median_duration(function_name: str) -> float:
    return median(durations(function_name))

def max_duration(function_name: str) -> float:
    return max(durations(function_name))

def _durations_dataframe() -> DataFrame:
    dataframe = DataFrame([
        {
            'function': function_name,
            'average': average_duration(function_name),
            'median': median_duration(function_name),
            'max': max_duration(function_name),
            'ct': len(durations(function_name))
        } for function_name in function_durations
    ])
    return dataframe.sort_values(by='average',ascending=False)


def export_durations() -> None:
    _durations_dataframe().to_csv(f'debug/{chain.id}/function_durations.csv')
    

def record_duration(function_name: str, inputs: str, duration: float) -> None:
    global num_recorded_durations
    event = {
        'inputs': inputs,
        'duration': duration,
    }
    function_durations[function_name].append(event)
    num_recorded_durations += 1
    if num_recorded_durations % 1000 == 0:
        export_durations()




import logging
import os
from random import randrange
from sqlite3 import OperationalError
from time import sleep, time
from typing import Any, Callable

from requests.exceptions import HTTPError, ReadTimeout

from y.utils.debug import record_duration

retry_logger = logging.getLogger('auto_retry')
DEBUG = os.environ.get('YPRICEMAGIC_DEBUG', False)

def continue_on_revert(func: Callable) -> Any:
    '''
    Decorates a call-making function. If the call reverts, returns `None`. Raises any other exceptions.
    '''
    from y.exceptions import continue_if_call_reverted

    def continue_on_revert_wrap(*args: Any, **kwargs: Any) -> Callable:
        try:
            return func(*args,**kwargs)
        except Exception as e:
            continue_if_call_reverted(e)
    
    return continue_on_revert_wrap


def log(logger: logging.Logger):
    """
    Decorates a function so both the inputs and outputs are logged with logger level DEBUG.
    For convenience, also decorates the function with @auto_retry.
    """

    def log_decorator(func: Callable) -> Callable:
        assert logger, 'To use @debug_logging decorator, you must pass in a logger.'

        def logging_wrap(*args: Any, **kwargs: Any) -> Any:
            fn_name = func.__name__
            
            if DEBUG:
                start = time()

            if len(kwargs) == 0:
                describer_string = f'{fn_name}{tuple([*args])}'
            else:
                describer_string = f'{fn_name}{tuple([*args])}, kwargs: {[*kwargs.items()]}'
            
            logger.debug(f'Fetching {describer_string}')
            func_returns = retry_superwrap(*args,**kwargs)
            logger.debug(f'{describer_string} returns: {func_returns}')

            # record function duration for debug purposes
            if DEBUG:
                record_duration(fn_name, describer_string, time() - start)
                
            return func_returns
        
        @auto_retry
        def retry_superwrap(*args: Any, **kwargs: Any) -> Callable:
            return func(*args, **kwargs)

        return logging_wrap
    return log_decorator

def auto_retry(func):
    '''
    Decorator that will retry the function on:
    - ConnectionError
    - HTTPError
    - TimeoutError
    - ReadTimeout
    
    It will also retry on specific ValueError exceptions:
    - Max rate limit reached
    - please use API Key for higher rate limit
    - execution aborted (timeout = 5s)
    
    On repeat errors, will retry in increasing intervals.
    '''

    def retry_wrap(*args, **kwargs):
        i = 0
        sleep_time = randrange(10,20)
        while True:
            try:
                return func(*args, **kwargs)
            except ValueError as e:
                retry_on_errs = (
                    # Occurs on any chain when making computationally intensive calls. Just retry.
                    'execution aborted (timeout = 5s)',

                    # Usually from block explorer while fetching contract source code. Just retry.
                    'Max rate limit reached',
                    'please use API Key for higher rate limit',

                    # Occurs occasionally on AVAX when node is slow to sync. Just retry.
                    'after last accepted block',
                )
                if 1 > 10 or not any([err in str(e) for err in retry_on_errs]):
                    raise
                retry_logger.warning(f'{str(e)} [{i}]')
            except (ConnectionError, HTTPError, TimeoutError, ReadTimeout) as e:
                # This happens when we pass too large of a request to the node. Do not retry.
                if 'Too Large' in str(e):
                    raise
                retry_logger.warning(f'{str(e)} [{i}]')
            except OperationalError as e:
                # This happens when brownie's deployments.db gets locked. Just retry.
                if 'database is locked' not in str(e):
                    raise
                retry_logger.warning(f'{str(e)} [{i}]')
            i += 1
            sleep(i * sleep_time)

    return retry_wrap



import logging
from functools import lru_cache
from typing import Optional, Tuple

from multicall import Call, Multicall
from y import convert
from y.classes.common import WeiBalance
from y.contracts import has_methods
from y.datatypes import UsdPrice, UsdValue
from y.decorators import log
from y.typing import AnyAddressType, Block
from y.utils.raw_calls import _totalSupplyReadable

logger = logging.getLogger(__name__)

@log(logger)
@lru_cache
def is_popsicle_lp(token_address: AnyAddressType) -> bool:
    # NOTE: contract to check for reference (mainnet): 0xd2C5A739ebfE3E00CFa88A51749d367d7c496CCf
    return has_methods(token_address, ['token0()(address)','token1()(address)','usersAmounts()((uint,uint))'])

@log(logger)
def get_price(token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdPrice]:
    address = convert.to_address(token)
    total_val = get_tvl(address, block)
    if total_val is None:
        return None
    total_supply = _totalSupplyReadable(address,block)
    return UsdPrice(total_val / total_supply)

@log(logger)
def get_tvl(token: AnyAddressType, block: Optional[Block] = None) -> Optional[UsdValue]:
    balances = get_balances(token, block)
    if balances is None:
        return None
    balance0, balance1 = balances
    return UsdValue(balance0.value_usd() + balance1.value_usd())

@log(logger)
def get_balances(token: AnyAddressType, block: Optional[Block] = None) -> Optional[Tuple[WeiBalance,WeiBalance]]:
    address = convert.to_address(token)
    methods = 'token0()(address)','token1()(address)','usersAmounts()((uint,uint))'
    calls = [Call(address, method, [[method,None]]) for method in methods]
    try:
        token0, token1, (balance0, balance1) = Multicall(calls, block_id=block)().values()
    except ValueError as e:
        if str(e) == "not enough values to unpack (expected 3, got 2)":
            # TODO determine if this is regular behavior when no tvl in pool or if this is bug to fix
            return None
        raise
    balance0 = WeiBalance(balance0, token0, block=block)
    balance1 = WeiBalance(balance1, token1, block=block)
    return balance0, balance1
